#! /usr/bin/env sh
# main module

# ╭──────╮
# │ main │
# ╰──────╯

# ╭──────┬───────────╮
# │ main │ constants │
# ╰──────┴───────────╯

# name of the entrypoint file
RWX_MAIN_NAME="main.sh"
RWX_SELF_NAME="rwx"

RWX_SELF_COMMAND="_${RWX_SELF_NAME}_cmd_"

# ╭──────┬───────────╮
# │ main │ variables │
# ╰──────┴───────────╯

RWX_COMMAND_ARGUMENT="${0}"
RWX_SHELL="$(cat "/proc/${$}/comm")"

RWX_COMMAND_NAME="$(basename "${RWX_COMMAND_ARGUMENT}" |
	sed "s|^-||")"
case "${RWX_COMMAND_NAME}" in
"bash" | "dash" | "sh") unset RWX_COMMAND_NAME ;;
*) ;;
esac
RWX_ROOT_SYSTEM="/usr/local/lib/${RWX_SELF_NAME}"
RWX_SELF_USER="${HOME}/${RWX_SELF_NAME}"

RWX_MAIN_PATH="${RWX_ROOT_SYSTEM}/${RWX_MAIN_NAME}"

# ╭──────┬──────╮
# │ main │ main │
# ╰──────┴──────╯

# run initial steps
rwx_main() {
	# cache main
	rwx_cache "${RWX_ROOT_SYSTEM}" "${RWX_MAIN_NAME}"
	# source system root
	if ! rwx_source "${RWX_ROOT_SYSTEM}" "${RWX_MAIN_NAME}"; then
		__rwx_log "Not a directory: ${RWX_ROOT_SYSTEM}"
		return 1
	fi
	# source user root
	rwx_source "${RWX_SELF_USER}"
	# parse code cache
	rwx_parse_code
	# context / command
	if [ -n "${RWX_COMMAND_NAME}" ]; then
		"${RWX_SELF_COMMAND}${RWX_COMMAND_NAME}" "${@}"
	# context / shell
	else
		rwx_self_init
	fi
}

# ╭──────┬───────╮
# │ main │ cache │
# ╰──────┴───────╯

# cache source code of a module
rwx_cache() {
	local root="${1}"
	local module="${2}"
	local name="${module%.sh}"
	local path="${root}/${module}"
	local text
	text="$(cat "${path}")"
	# all source code
	RWX_CODE="${RWX_CODE}\
#. ${name}
${text}
"
}

# ╭──────┬────────╮
# │ main │ source │
# ╰──────┴────────╯

# source code from file path
rwx_source() {
	local root="${1}"
	[ -d "${root}" ] ||
		return 1
	local file="${2}"
	local count module
	count=0
	__rwx_log "" \
		". ${root}"
	local modules="$(rwx_find_shell "${root}" "${file}")"
	while IFS= read -r module; do
		count=$((count + 1))
		__rwx_log "$(printf "%02d" "${count}") ${module%.sh}"
		# shellcheck disable=SC1090
		. "${root}/${module}"
		# cache code
		rwx_cache "${root}" "${module}"
	done <<EOF
${modules}
EOF
}

# ╭──────┬─────╮
# │ main │ log │
# ╰──────┴─────╯

__rwx_log() {
	if rwx_shell_interactive; then
		[ ${#} -gt 0 ] || set -- ""
		local line
		for line in "${@}"; do
			echo "${line}"
		done
	fi
}

# ╭──────┬───────╮
# │ main │ shell │
# ╰──────┴───────╯

# test if active shell is in interactive mode
rwx_shell_interactive() {
	case "${-}" in
	*i*) ;;
	*) return 1 ;;
	esac
}

# ╭──────┬──────╮
# │ main │ find │
# ╰──────┴──────╯

# find directory’s files by extension
rwx_find_extension() {
	local extension="${1}"
	local root="${2}"
	local file="${3}"
	set -- \
		"${root}" \
		-name "*.${extension}" \
		-type "f"
	[ -n "${file}" ] &&
		set -- "${@}" \
			-not \
			-name "${file}"
	find "${@}" \
		-printf "%P\n" |
		sort
}

# find directory’s sh files
rwx_find_shell() {
	rwx_find_extension "sh" "${@}"
}

# ╭──────┬───────╮
# │ main │ parse │
# ╰──────┴───────╯

RWX_REGEX_CONSTANT="[_A-Z][_0-9A-Z]*"
RWX_REGEX_FUNCTION="[_A-Za-z][_0-9A-Za-z]*"
RWX_REGEX_SET="=.*"
RWX_REGEX_SPACES="[[:space:]]*"

RWX_REGEX_BEGIN="^${RWX_REGEX_SPACES}"
RWX_REGEX_OPEN="\
${RWX_REGEX_SPACES}(${RWX_REGEX_SPACES})${RWX_REGEX_SPACES}{.*"

RWX_REGEX_TARGET_CONSTANT="\
${RWX_REGEX_BEGIN}\\(${RWX_REGEX_CONSTANT}\\)${RWX_REGEX_SET}"
RWX_REGEX_TARGET_DOC="\
${RWX_REGEX_BEGIN}# \\(.*\\)${RWX_REGEX_SPACES}\$"
RWX_REGEX_TARGET_FUNCTION="\
${RWX_REGEX_BEGIN}\\(${RWX_REGEX_FUNCTION}\\)${RWX_REGEX_OPEN}"

rwx_parse_code() {
	# parse constants
	RWX_CONSTANTS="$(rwx_parse_constants)"
	# parse functions
	RWX_FUNCTIONS="$(rwx_parse_functions)"
}
rwx_parse_constants() {
	printf "%s\n" "${RWX_CODE}" |
		sed "/${RWX_REGEX_TARGET_CONSTANT}/s|${RWX_REGEX_TARGET_CONSTANT}|\\1|"
}
rwx_parse_functions() {
	printf "%s\n" "${RWX_CODE}" |
		sed "/${RWX_REGEX_TARGET_FUNCTION}/s|${RWX_REGEX_TARGET_FUNCTION}|\\1|"
}

# ╭──────┬─────╮
# │ main │ run │
# ╰──────┴─────╯

# run main function
rwx_main "${@}"
